import json

from agave.tools.sync.sqs_client import SqsClient

CORE_QUEUE_REGION = 'us-east-1'


async def test_send_message(sqs_client) -> None:
    data1 = dict(hola='mundo')
    data2 = dict(foo='bar')

    client = SqsClient(sqs_client.queue_url, CORE_QUEUE_REGION)
    client.send_message(data1)
    client.send_message(data2, message_group_id='12345')

    sqs_message = await sqs_client.receive_message()
    message_body = json.loads(sqs_message['Messages'][0]['Body'])
    assert message_body == data1

    sqs_message = await sqs_client.receive_message()
    message = json.loads(sqs_message['Messages'][0]['Body'])
    assert message == data2



import json
from typing import Annotated, Union
from unittest.mock import AsyncMock, call

from cuenca_validations.types import LogConfig
from pydantic import BaseModel

from agave.core.exc import RetryTask
from agave.tasks.sqs_tasks import task

from ..utils import CORE_QUEUE_REGION, extract_log_data


async def test_execute_tasks_logger(sqs_client, caplog) -> None:
    test_message = dict(foo='bar')
    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async_mock_function = AsyncMock()

    async def my_task_with_logger(data: dict) -> None:
        await async_mock_function(data)

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(my_task_with_logger)()
    async_mock_function.assert_called_with(test_message)

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert len(log_data) == 1  # one log entry for the task
    assert log_data[0]['request']['task_func'] == my_task_with_logger.__name__
    assert log_data[0]['request']['body'] == test_message
    assert log_data[0]['response']['status'] == 'success'
    assert log_data[0]['response']['body'] is None


async def test_execute_tasks_with_validator_logger(sqs_client, caplog) -> None:
    class ValidatorRequest(BaseModel):
        id: str
        secret: Annotated[str, LogConfig(masked=True, unmasked_chars_length=3)]

    class ValidatorResponse(BaseModel):
        id: str
        api_key: Annotated[
            str, LogConfig(masked=True, unmasked_chars_length=3)
        ]
        status: str

    async_mock_function = AsyncMock(return_value=None)

    async def my_task(data: ValidatorRequest) -> ValidatorResponse:
        await async_mock_function(data)
        return ValidatorResponse(
            id='abc123', api_key='1234567890', status='success'
        )

    data = dict(id='abc123', secret='my-secret')
    expected_message = dict(id='abc123', secret='*****ret')
    expected_response = dict(id='abc123', api_key='*****890', status='success')

    test_message = ValidatorRequest(**data)
    await sqs_client.send_message(
        MessageBody=test_message.model_dump_json(),
        MessageGroupId='1234',
    )
    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(my_task)()
    async_mock_function.assert_called_with(test_message)

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['task_func'] == my_task.__name__
    assert log_data[0]['request']['body'] == expected_message
    assert log_data[0]['response']['status'] == 'success'
    assert log_data[0]['response']['body'] == expected_response


async def test_execute_tasks_with_union_validator_logger(
    sqs_client, caplog
) -> None:
    class User(BaseModel):
        id: str
        name: str
        secret: Annotated[str, LogConfig(masked=True, unmasked_chars_length=3)]

    class Company(BaseModel):
        id: str
        legal_name: str
        rfc: str
        secret: Annotated[str, LogConfig(masked=True, unmasked_chars_length=3)]

    async_mock_function = AsyncMock(return_value=None)

    async def my_task(data: Union[User, Company]) -> None:
        await async_mock_function(data.model_dump())

    task_params = dict(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )
    test_message = dict(
        id='ID123', name='Sor Juana Inés de la Cruz', secret='my-secret'
    )
    expected_message_user = dict(
        id='ID123', name='Sor Juana Inés de la Cruz', secret='*****ret'
    )
    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='4321',
    )
    await task(**task_params)(my_task)()
    async_mock_function.assert_called_with(test_message)

    async_mock_function.reset_mock()
    test_message = dict(
        id='ID123',
        legal_name='FastAgave',
        rfc='FA123456789',
        secret='my-secret',
    )
    expected_message_company = dict(
        id='ID123',
        legal_name='FastAgave',
        rfc='FA123456789',
        secret='*****ret',
    )
    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='54321',
    )
    await task(**task_params)(my_task)()
    async_mock_function.assert_called_with(test_message)

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert len(log_data) == 2  # two log entries for the task
    assert log_data[0]['request']['body'] == expected_message_user
    assert log_data[1]['request']['body'] == expected_message_company


async def test_execute_tasks_with_response_logger(sqs_client, caplog) -> None:
    test_message = dict(foo='bar')

    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )
    async_mock_function = AsyncMock()

    async def my_task_with_logger(data: dict) -> dict:
        await async_mock_function(data)
        my_response = dict(response='my_custom_response')
        return my_response

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(my_task_with_logger)()
    async_mock_function.assert_called_with(test_message)

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['task_func'] == my_task_with_logger.__name__
    assert log_data[0]['request']['body'] == test_message
    assert log_data[0]['response']['status'] == 'success'
    assert log_data[0]['response']['body'] == dict(
        response='my_custom_response'
    )


async def test_execute_tasks_with_response_non_serializable(
    sqs_client, caplog
) -> None:

    class NonSerializableResponse:
        def __init__(self, value):
            self.value = value

    test_message = dict(foo='bar')

    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )
    async_mock_function = AsyncMock()

    async def my_task(data: dict) -> NonSerializableResponse:
        await async_mock_function(data)
        return NonSerializableResponse('test_value')

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(my_task)()
    async_mock_function.assert_called_with(test_message)

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['response']['status'] == 'success'
    assert (
        'NonSerializableResponse object at' in log_data[0]['response']['body']
    )


async def test_retry_tasks_default_max_retries_logger(
    sqs_client, caplog
) -> None:
    test_message = dict(id='abc123', name='fast-agave')

    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async_mock_function = AsyncMock(side_effect=RetryTask)

    async def my_task(data: dict) -> None:
        await async_mock_function(data)

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
        max_retries=2,
    )(my_task)()

    expected_calls = [call(test_message)] * 3
    async_mock_function.assert_has_calls(expected_calls)
    assert async_mock_function.call_count == len(expected_calls)

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert len(log_data) == 3  # 3 log entries for the task

    assert log_data[0]['request']['task_func'] == my_task.__name__
    assert log_data[0]['request']['body'] == test_message

    # For the first execution
    assert log_data[0]['response']['status'] == 'retrying'
    assert (
        log_data[0]['request']['message_attributes']['ApproximateReceiveCount']
        == '1'
    )

    # For the second execution
    assert log_data[1]['response']['status'] == 'retrying'
    assert (
        log_data[1]['request']['message_attributes']['ApproximateReceiveCount']
        == '2'
    )

    # For the third execution
    assert log_data[2]['response']['status'] == 'retrying'
    assert (
        log_data[2]['request']['message_attributes']['ApproximateReceiveCount']
        == '3'
    )


async def test_task_with_exception_logger(sqs_client, caplog) -> None:
    test_message = dict(foo='bar')
    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async def my_task_with_logger(data: dict) -> None:
        raise Exception('test_exception')

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
        max_retries=2,
    )(my_task_with_logger)()

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['response']['status'] == 'failed'
    assert log_data[0]['response']['error'] == 'test_exception'



from typing import Optional, Union

import pytest
from cuenca_validations.types.general import LogConfig
from pydantic import BaseModel

from agave.core.loggers import (
    get_request_model,
    get_response_model,
    obfuscate_sensitive_data,
)


@pytest.mark.parametrize(
    "body, sensitive_fields, expected_result",
    [
        # Test 1: Obfuscate with exclusion
        (
            {"username": "user123", "password": "secret"},
            {"password": LogConfig(excluded=True)},
            {"username": "user123"},
        ),
        # Test 2: Obfuscate with masking
        (
            {"username": "user123", "password": "secret"},
            {"password": LogConfig(masked=True)},
            {"username": "user123", "password": "*****"},
        ),
        # Test 3: Obfuscate with partial masking
        (
            {"username": "user123", "password": "supersecret"},
            {"password": LogConfig(masked=True, unmasked_chars_length=4)},
            {"username": "user123", "password": "*****cret"},
        ),
        # Test 4: Obfuscate non-existing field
        (
            {"username": "user123"},
            {"password": LogConfig(masked=True)},
            {"username": "user123"},
        ),
    ],
)
def test_obfuscate_sensitive_data(body, sensitive_fields, expected_result):
    result = obfuscate_sensitive_data(body, sensitive_fields)
    assert result == expected_result


@pytest.mark.parametrize(
    'param_type, response_type',
    [
        (int, int),
        (float, float),
        (str, str),
        (bool, bool),
        (dict, dict),
        (Optional[int], Optional[int]),
        (list[int], list[int]),
        (tuple[int, str], tuple[int, str]),
        (set[int], set[int]),
        (Union[list[int], str], str),
    ],
)
def test_get_request_model_invalid_types(param_type, response_type) -> None:
    def my_function(param: param_type) -> response_type:
        return param

    request_model = get_request_model(my_function)
    response_model = get_response_model(my_function)

    assert request_model is None
    assert response_model is None


def test_get_request_model_valid_types() -> None:
    class RequestModel(BaseModel):
        id: int
        name: str

    class ResponseModel(BaseModel):
        id: int

    def my_function(param: RequestModel) -> ResponseModel:
        return ResponseModel(id=param.id)

    request_model = get_request_model(my_function)
    response_model = get_response_model(my_function)

    assert request_model == [RequestModel]
    assert response_model == ResponseModel



import logging
from tempfile import TemporaryFile

import pytest
from fastapi.testclient import TestClient

from examples.models import Account

from ..utils import extract_log_data


def test_logger_no_request_model(fastapi_client: TestClient, caplog) -> None:
    """
    Test that verifies:
    - A resource can be retrieved successfully (HTTP 200)
    - The request and response are properly logged
    """
    caplog.set_level(logging.INFO)

    request_data = {
        'some_field': 'some value',
    }
    response = fastapi_client.post('/token', json=request_data)
    response_body = response.json()

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['body'] == request_data
    assert log_data[0]['response']['body'] == response_body


def test_logger_retrieve_resource(
    fastapi_client: TestClient, account: Account, caplog
) -> None:
    """
    Test that verifies:
    - A resource can be retrieved successfully (HTTP 200)
    - The request and response are properly logged
    """
    response = fastapi_client.get(f'/accounts/{account.id}')
    response_body = response.json()

    assert response.status_code == 200
    assert response_body == account.to_dict()

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)
    assert log_data[0]['request']['url'].endswith(f'/accounts/{account.id}')
    assert log_data[0]['response']['status_code'] == 200
    assert log_data[0]['response']['body']['name'] == '*****Kahlo'


def test_logger_update_resource(
    fastapi_client: TestClient, account: Account, caplog
) -> None:
    """
    Test that verifies:
    - A resource can be updated successfully (HTTP 200)
    - The request and response are properly logged
    """
    resp = fastapi_client.patch(
        f'/accounts/{account.id}',
        json=dict(name='Maria Felix'),
    )
    json_body = resp.json()
    status_code = resp.status_code
    account.reload()
    assert json_body['name'] == 'Maria Felix'
    assert status_code == 200

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['response']['status_code'] == 200
    assert log_data[0]['response']['body']['name'] == '*****Felix'


def test_logger_create_resource_bad_request(
    fastapi_client: TestClient, caplog
) -> None:
    """
    Test that verifies the logger correctly captures and logs a 422 Bad Request
    when an invalid request payload is sent.
    """
    request_data = {'invalid_field': 'some value'}
    response = fastapi_client.post('/accounts', json=request_data)

    assert response.status_code == 422

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['body'] == request_data
    assert log_data[0]['response']['status_code'] == 422


def test_logger_test_logger_retrieve_resource_not_found(
    fastapi_client: TestClient, caplog
) -> None:
    """
    Test that verifies:
    - A request for a non-existent resource returns a 404
    - The request and response are properly logged
    """
    resource_id = "unknown_id"
    response = fastapi_client.get(f"/accounts/{resource_id}")

    assert response.status_code == 404

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['url'].endswith(f"/accounts/{resource_id}")
    assert log_data[0]['response']['status_code'] == 404


def test_logger_upload_resource(fastapi_client: TestClient, caplog) -> None:
    """
    Test that verifies the logger properly handles file uploads, ensuring:
    - Request body is logged as None (since files are not logged)
    - Content-Type is correctly set to multipart/form-data
    """
    with TemporaryFile(mode='rb') as f:
        file_body = f.read()
    response = fastapi_client.post(
        '/files',
        files=dict(file=(None, file_body), file_name=(None, 'test_file.txt')),
    )

    assert response.status_code == 201
    response_body = response.json()
    assert response_body['name'] == 'test_file.txt'

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['body'] is None
    assert log_data[0]['request']['headers']['content-type'].startswith(
        'multipart/form-data'
    )


def test_logger_headers_case_insensitive(
    fastapi_client: TestClient, caplog
) -> None:
    request_data = {
        'user': 'user',
        'password': 'My-super-secret-password',
        'short_secret': '123',
    }

    request_headers = {
        'x-cuenca-loginid': 'My-secret-login-id',
        'X-CUENCA-LOGINTOKEN': 'My-secret-login-token',
        'AUTHORIZATION': '123',
        'Content-Type': 'application/json',
        'connection': 'keep-alive',
    }

    response = fastapi_client.post(
        '/api_keys', json=request_data, headers=request_headers
    )

    assert response.status_code == 201

    expected_log_headers = {
        'host': 'testserver',
        'accept': '*/*',
        'accept-encoding': 'gzip, deflate',
        'user-agent': 'testclient',
        'x-cuenca-loginid': '*****n-id',
        'x-cuenca-logintoken': '*****oken',
        'authorization': '*****',
        'content-type': 'application/json',
    }

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['headers'] == expected_log_headers


def test_logger_api_route(fastapi_client: TestClient, caplog) -> None:
    """
    Test that verifies the logger properly masks sensitive data in:
    - Request headers
    - Request body fields
    - Response body fields
    """
    request_data = {
        'user': 'user',
        'password': 'My-super-secret-password',
        'short_secret': '123',
    }
    request_headers = {
        'X-Cuenca-LoginId': 'My-secret-login-id',
        'X-Cuenca-LoginToken': 'My-secret-login-token',
        'Authorization': '123',
        'Content-Type': 'application/json',
        'Connection': 'keep-alive',
    }

    response = fastapi_client.post(
        '/api_keys', json=request_data, headers=request_headers
    )
    response_body = response.json()

    assert response.status_code == 201
    assert response_body['secret'] == 'My-super-secret-key'
    assert response_body['password'] == 'My-super-secret-password'

    expected_log = {
        'request': {
            'method': 'POST',
            'url': 'http://testserver/api_keys',
            'query_params': '',
            'headers': {
                'host': 'testserver',
                'accept': '*/*',
                'accept-encoding': 'gzip, deflate',
                'user-agent': 'testclient',
                'x-cuenca-loginid': '*****n-id',
                'x-cuenca-logintoken': '*****oken',
                'authorization': '*****',
                'content-type': 'application/json',
            },
            'body': {
                'user': 'user',
                'password': '*****',
                'short_secret': '*****',
            },
        },
        'response': {
            'status_code': 201,
            'headers': {'content-type': 'application/json'},
            'body': {
                'id': response_body['id'],
                'secret': '*****-key',
                'user': 'user',
                'password': '*****word',
                'user_id': 'US123456789',
                'platform_id': 'PT123456',
                'created_at': None,
                'another_field': '12345678',
            },
        },
    }

    # Extract and validate logger output
    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0] == expected_log


def test_logger_internal_server_error(
    fastapi_client: TestClient, caplog
) -> None:
    """
    Test that verifies:
    - A request that causes a 500 error is logged.
    """
    request_data = {'some_field': 'some value'}

    with pytest.raises(Exception):
        fastapi_client.post('/simulate_500', json=request_data)

    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['body'] == request_data
    assert log_data[0]['response']['status_code'] == 500


def test_logger_bad_request(fastapi_client: TestClient, caplog) -> None:
    """
    Test that verifies:
    - A request that causes a 400 error is logged.
    """
    request_data = {'some_field': 'some value'}
    response = fastapi_client.post('/simulate_400', json=request_data)
    assert response.status_code == 400

    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['body'] == request_data


def test_logger_unauthorized(fastapi_client: TestClient, caplog) -> None:
    """
    Test that verifies:
    - A request that causes a 401 error is logged.
    """
    request_data = {'some_field': 'some value'}
    response = fastapi_client.post('/simulate_401', json=request_data)
    assert response.status_code == 401

    log_output = caplog.text
    log_data = extract_log_data(log_output)

    assert log_data[0]['request']['body'] == request_data



import json
import re

from chalice.test import Client as OriginalChaliceClient

CORE_QUEUE_REGION = 'us-east-1'


def extract_log_data(log_output: str) -> list[dict]:
    """
    Extracts JSON data from log output using a regex
    pattern and returns a list of matches.
    """
    matches = re.findall(r"(\{.*\})", log_output)
    if not matches:
        return []

    return [json.loads(match) for match in matches]


class ChaliceResponse:
    def __init__(self, chalice_response):
        self._response = chalice_response
        self._json_body = chalice_response.json_body
        self._status_code = chalice_response.status_code
        self._headers = chalice_response.headers

    def json(self):
        return self._json_body

    @property
    def status_code(self):
        return self._status_code

    @property
    def headers(self):
        return self._headers


class ChaliceClient(OriginalChaliceClient):
    def _request_with_json(
        self, method: str, url: str, **kwargs
    ) -> ChaliceResponse:
        body = json.dumps(kwargs.pop('json')) if 'json' in kwargs else None
        headers = {'Content-Type': 'application/json'}
        response = getattr(self.http, method)(
            url, body=body, headers=headers, **kwargs
        )
        return ChaliceResponse(response)

    def post(self, url: str, **kwargs) -> ChaliceResponse:
        return self._request_with_json('post', url, **kwargs)

    def get(self, url: str, **kwargs) -> ChaliceResponse:
        response = self.http.get(url, **kwargs)
        return ChaliceResponse(response)

    def patch(self, url: str, **kwargs) -> ChaliceResponse:
        return self._request_with_json('patch', url, **kwargs)

    def delete(self, url: str, **kwargs) -> ChaliceResponse:
        return self._request_with_json('delete', url, **kwargs)



from dataclasses import dataclass
from typing import Optional


@dataclass
class AgaveError(Exception):
    error: str
    status_code: int


@dataclass
class BadRequestError(AgaveError):
    status_code: int = 400


@dataclass
class UnauthorizedError(AgaveError):
    status_code: int = 401


@dataclass
class ForbiddenError(AgaveError):
    status_code: int = 403


@dataclass
class NotFoundError(AgaveError):
    status_code: int = 404


@dataclass
class MethodNotAllowedError(AgaveError):
    status_code: int = 405


@dataclass
class ConflictError(AgaveError):
    status_code: int = 409


@dataclass
class UnprocessableEntity(AgaveError):
    status_code: int = 422


@dataclass
class TooManyRequests(AgaveError):
    status_code: int = 429


@dataclass
class AgaveViewError(AgaveError):
    status_code: int = 500


@dataclass
class ServiceUnavailableError(AgaveError):
    status_code: int = 503


@dataclass
class RetryTask(Exception):
    countdown: Optional[int] = None



import json
from dataclasses import dataclass, field
from typing import Optional, Union
from uuid import uuid4

try:
    import boto3
    from types_boto3_sqs import SQSClient as Boto3SQSClient
except ImportError:
    raise ImportError(
        "You must install agave with [sync_aws_tools] option.\n"
        "You can install it with: pip install agave[sync_aws_tools]"
    )


@dataclass
class SqsClient:
    queue_url: str
    region_name: str
    _sqs: Boto3SQSClient = field(init=False)

    def __post_init__(self) -> None:
        self._sqs = boto3.client('sqs', region_name=self.region_name)

    def send_message(
        self,
        data: Union[str, dict],
        message_group_id: Optional[str] = None,
    ) -> None:
        self._sqs.send_message(
            QueueUrl=self.queue_url,
            MessageBody=data if isinstance(data, str) else json.dumps(data),
            MessageGroupId=message_group_id or str(uuid4()),
        )



import importlib
import sys

import pytest


@pytest.mark.parametrize(
    'module_path,required_types,required_option',
    [
        (
            'agave.tools.asyncio.sqs_client',
            'types_aiobotocore_sqs',
            'asyncio_aws_tools',
        ),
        (
            'agave.tools.sync.sqs_client',
            'types_boto3_sqs',
            'sync_aws_tools',
        ),
    ],
)
def test_tasks_import_error(
    monkeypatch: pytest.MonkeyPatch,
    module_path: str,
    required_types: str,
    required_option: str,
) -> None:
    # Clear modules from sys.modules
    for module in [required_types, module_path]:
        if module in sys.modules:
            del sys.modules[module]

    monkeypatch.setitem(sys.modules, required_types, None)

    with pytest.raises(ImportError) as exc_info:
        importlib.import_module(module_path)

    assert f"You must install agave with [{required_option}] option" in str(
        exc_info.value
    )
    assert f"pip install agave[{required_option}]" in str(exc_info.value)



import importlib
import sys

import pytest


def test_fast_import_error(monkeypatch):
    for module in ['fastapi', 'agave.fastapi.rest_api']:
        if module in sys.modules:
            del sys.modules[module]

    monkeypatch.setitem(sys.modules, 'fastapi', None)

    with pytest.raises(ImportError) as exc_info:
        importlib.import_module('agave.fastapi.rest_api')

    assert "You must install agave with [fastapi] option" in str(
        exc_info.value
    )
    assert "pip install agave[fastapi]" in str(exc_info.value)



import uuid
from base64 import urlsafe_b64encode


# This function is used to generate an id composed of a
# list of fields in alphabetical order, for example if we want
# uuid_field_generic('AC', account_number='bla', user_id='ble')
# it will always generate the same id, because it is based on the arguments
def uuid_field_generic(prefix: str = '', **fields) -> str:
    sorted_text = ''.join(
        [
            item[1]
            for item in sorted(fields.items(), key=lambda x: x[0].lower())
        ]
    )
    uuid5 = uuid.uuid5(uuid.NAMESPACE_OID, sorted_text)
    return prefix + urlsafe_b64encode(uuid5.bytes).decode()[:-2]



import json
from base64 import b64encode
from typing import Iterable
from uuid import uuid4


def _b64_encode(value: str) -> str:
    encoded = b64encode(bytes(value, 'utf-8'))
    return encoded.decode('utf-8')


def build_celery_message(
    task_name: str, args_: Iterable, kwargs_: dict
) -> str:
    task_id = str(uuid4())
    # la definición de esta plantila se encuentra en:
    # docs.celeryproject.org/en/stable/internals/protocol.html#definition
    message = dict(
        properties=dict(
            correlation_id=task_id,
            content_type='application/json',
            content_encoding='utf-8',
            body_encoding='base64',
            delivery_info=dict(exchange='', routing_key='celery'),
        ),
        headers=dict(
            lang='py',
            task=task_name,
            id=task_id,
            root_id=task_id,
            parent_id=None,
            group=None,
        ),
        body=_b64_encode(
            json.dumps(
                (
                    args_,
                    kwargs_,
                    dict(
                        callbacks=None, errbacks=None, chain=None, chord=None
                    ),
                )
            )
        ),
    )
    message['content-encoding'] = 'utf-8'
    message['content-type'] = 'application/json'

    encoded = _b64_encode(json.dumps(message))
    return encoded



import uuid
from base64 import urlsafe_b64encode

from agave.chalice.models.helpers import uuid_field_generic


def test_uuid_field_generic():
    generic_asc = uuid_field_generic(
        'AC',
        account_number='123456789',
        bank_code='90464',
        user_id='US01',
    )

    uuid5 = uuid.uuid5(uuid.NAMESPACE_OID, '12345678990464US01')
    assert generic_asc == 'AC' + urlsafe_b64encode(uuid5.bytes).decode()[:-2]

    generic_desc = uuid_field_generic(
        'AC',
        user_id='US01',
        bank_code='90464',
        account_number='123456789',
    )

    assert generic_asc == generic_desc



from typing import (
    Any,
    Callable,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from cuenca_validations.types.general import LogConfig
from cuenca_validations.types.helpers import get_log_config
from pydantic import BaseModel

HEADERS_LOG_CONFIG = {
    'authorization': LogConfig(masked=True),
    'x-cuenca-token': LogConfig(masked=True, unmasked_chars_length=4),
    'x-cuenca-loginid': LogConfig(masked=True, unmasked_chars_length=4),
    'x-cuenca-logintoken': LogConfig(masked=True, unmasked_chars_length=4),
    'x-cuenca-sessionid': LogConfig(masked=True, unmasked_chars_length=4),
    'connection': LogConfig(excluded=True),
    'content-length': LogConfig(excluded=True),
}


def obfuscate_sensitive_data(
    body: dict[str, Any],
    sensitive_fields: dict[str, LogConfig],
) -> dict[str, Any]:

    ofuscated_body = body.copy()
    for field_name, log_config in sensitive_fields.items():
        if field_name not in ofuscated_body:
            continue

        if log_config.excluded:
            del ofuscated_body[field_name]
        else:
            value = ofuscated_body[field_name]
            log_chars = log_config.unmasked_chars_length
            ofuscated_body[field_name] = (
                '*****' + value[-log_chars:] if log_chars > 0 else '*****'
            )

    return ofuscated_body


def get_request_model(
    function: Callable[..., Any],
) -> Optional[list[Type[BaseModel]]]:
    """
    Extracts the first parameter from a function that is a
    BaseModel or Union of BaseModels.
    """
    type_hints = get_type_hints(function)
    type_hints.pop('return', None)

    extracted_types: list[type] = []
    valid_types: list[type] = []

    for param_type in type_hints.values():
        origin = get_origin(param_type)
        if origin is Union:
            extracted_types.extend(get_args(param_type))
        else:
            extracted_types.append(param_type)

    for model_type in extracted_types:
        try:
            if issubclass(model_type, BaseModel):
                valid_types.append(model_type)
        except TypeError:
            continue

    return valid_types or None


def get_response_model(
    function: Callable[..., Any]
) -> Optional[Type[BaseModel]]:
    """
    Extracts the response model from the function's return type
    if it is a Pydantic BaseModel.
    """
    hints = get_type_hints(function)
    return_annotation = hints.get('return', None)

    if return_annotation is None:
        return None

    try:
        if issubclass(return_annotation, BaseModel):
            return return_annotation
        else:
            return None
    except TypeError:
        return None


def get_sensitive_fields(
    models: Optional[Union[list[type[BaseModel]], type[BaseModel]]],
) -> dict[str, Any]:
    """
    Analyzes a list of Pydantic models and returns
    a set of field names marked as sensitive in their metadata.
    """
    sensitive_fields: dict[str, Any] = {}

    if models is None or models is Any:
        return sensitive_fields

    if not isinstance(models, list):
        models = [models]

    for m in models:
        for field_name, field in m.model_fields.items():
            log_config = get_log_config(field)
            if not log_config:
                continue
            if log_config.masked or log_config.excluded:
                sensitive_fields[field_name] = log_config

    return sensitive_fields



from importlib.machinery import SourceFileLoader

from setuptools import find_packages, setup

version = SourceFileLoader('version', 'agave/version.py').load_module()

with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='agave',
    version=version.__version__,
    author='Cuenca',
    author_email='dev@cuenca.com',
    description='Rest_api',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cuenca-mx/agave',
    packages=find_packages(),
    include_package_data=True,
    package_data=dict(agave=['py.typed']),
    python_requires='>=3.9',
    install_requires=[
        'cuenca-validations>=2.1.0,<3.0.0',
        'mongoengine>=0.29.0,<0.30.0',
        'mongoengine-plus>=1.0.0,<2.0.0',
        'python-multipart>=0.0.20,<0.0.30',
    ],
    extras_require={
        'chalice': [
            'chalice>=1.30.0,<2.0.0',
        ],
        'fastapi': [
            'fastapi>=0.115.0,<1.0.0',
            # TODO: Remove this once we upgrade to starlette:
            # This is a temporary dependency to avoid breaking changes.
            # https://github.com/cuenca-mx/agave/issues/158
            'starlette>=0.45.0,<0.46.0',
            'starlette-context>=0.3.2,<0.4.0',
        ],
        'tasks': [
            'aiobotocore>=2.0.0,<3.0.0',
            'types-aiobotocore-sqs>=2.1.0,<3.0.0',
        ],
        'sync_aws_tools': [
            'boto3>=1.34.106,<2.0.0',
            'types-boto3[sqs]>=1.34.106,<2.0.0',
        ],
        'asyncio_aws_tools': [
            'aiobotocore>=2.0.0,<3.0.0',
            'types-aiobotocore-sqs>=2.1.0,<3.0.0',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import json
import logging
from json import JSONDecodeError
from typing import Any, Callable

from cuenca_validations.errors import CuencaError
from fastapi import HTTPException, Request, Response
from fastapi.exceptions import RequestValidationError
from fastapi.routing import APIRoute

from ...core.exc import AgaveError
from ...core.loggers import HEADERS_LOG_CONFIG, obfuscate_sensitive_data

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)


def get_status_code_exception(exc: Exception) -> int:
    if isinstance(exc, HTTPException):
        return exc.status_code
    if isinstance(exc, RequestValidationError):
        return 422
    if isinstance(exc, AgaveError):
        return exc.status_code
    if isinstance(exc, CuencaError):
        return exc.status_code
    return 500


class LoggingRoute(APIRoute):

    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def logging_route_handler(request: Request) -> Response:

            req_handler = request.scope['route_handler']

            try:
                ofuscated_request_body = obfuscate_sensitive_data(
                    await request.json(),
                    req_handler.endpoint.request_log_config_fields,
                )
            except AttributeError:
                ofuscated_request_body = await request.json()
            except JSONDecodeError:
                ofuscated_request_body = None

            log_data: dict[str, Any] = {
                'request': {
                    'method': request.method,
                    'url': str(request.url),
                    'query_params': request.query_params,
                    'headers': obfuscate_sensitive_data(
                        dict(request.headers),
                        HEADERS_LOG_CONFIG,
                    ),
                    'body': ofuscated_request_body,
                }
            }

            response: Response

            try:
                response = await original_route_handler(request)
            except Exception as exc:
                log_data['response'] = {
                    'status_code': get_status_code_exception(exc)
                }
                raise
            else:

                if hasattr(response, 'body'):
                    try:
                        ofuscated_response_body = obfuscate_sensitive_data(
                            json.loads(response.body),
                            req_handler.endpoint.response_log_config_fields,
                        )
                    except AttributeError:
                        ofuscated_response_body = json.loads(response.body)
                else:
                    ofuscated_response_body = None

                log_data['response'] = {
                    'status_code': response.status_code,
                    'headers': obfuscate_sensitive_data(
                        dict(response.headers),
                        HEADERS_LOG_CONFIG,
                    ),
                    'body': ofuscated_response_body,
                }

            finally:
                logger.info(json.dumps(log_data, default=str))

            return response

        return logging_route_handler



import asyncio
from dataclasses import dataclass
from typing import Iterable, Optional

from ..celery import build_celery_message
from .sqs_client import SqsClient


@dataclass
class SqsCeleryClient(SqsClient):
    async def send_task(
        self,
        name: str,
        args: Optional[Iterable] = None,
        kwargs: Optional[dict] = None,
    ) -> None:
        celery_message = build_celery_message(name, args or (), kwargs or {})
        await super().send_message(celery_message)

    def send_background_task(
        self,
        name: str,
        args: Optional[Iterable] = None,
        kwargs: Optional[dict] = None,
    ) -> asyncio.Task:
        celery_message = build_celery_message(name, args or (), kwargs or {})
        return super().send_message_async(celery_message)



import datetime as dt
from tempfile import TemporaryFile
from unittest.mock import MagicMock, patch
from urllib.parse import urlencode

import pytest
from fastapi.testclient import TestClient

from examples.config import (
    TEST_DEFAULT_PLATFORM_ID,
    TEST_DEFAULT_USER_ID,
    TEST_SECOND_PLATFORM_ID,
)
from examples.models import Account, Card, File, User

# Constants for both frameworks
FRAMEWORK_CONFIGS = {
    'chalice': {
        'platform_id_filter': (
            'examples.chalice.blueprints.authed.'
            'AuthedBlueprint.platform_id_filter_required'
        ),
        'user_id_filter': (
            'examples.chalice.blueprints.authed.'
            'AuthedBlueprint.user_id_filter_required'
        ),
        'bad_request_code': 400,
        'validation_error_code': 400,
        'method_not_allowed_code': 400,
        'not_found_code': 403,
    },
    'fastapi': {
        'platform_id_filter': (
            'examples.fastapi.middlewares.'
            'AuthedMiddleware.required_platform_id'
        ),
        'user_id_filter': (
            'examples.fastapi.middlewares.AuthedMiddleware.required_user_id'
        ),
        'bad_request_code': 422,
        'validation_error_code': 422,
        'method_not_allowed_code': 405,
        'not_found_code': 404,
    },
}


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_create_resource(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    data = dict(name='Doroteo Arango')
    resp = client.post("/accounts", json=data)
    json_body = resp.json()
    status_code = resp.status_code
    model = Account.objects.get(id=json_body['id'])
    assert status_code == 201
    assert model.to_dict() == json_body
    model.delete()


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_create_resource_bad_request(
    client_fixture: str, framework_config: dict, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    data = dict(invalid_field='some value')
    resp = client.post('/accounts', json=data)
    assert resp.status_code == framework_config['validation_error_code']


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_retrieve_resource(
    client_fixture: str, request: pytest.FixtureRequest, account: Account
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.get(f'/accounts/{account.id}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert json_body == account.to_dict()


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_retrieve_resource_platform_id_filter_required(
    client_fixture: str,
    framework_config: dict,
    request: pytest.FixtureRequest,
    other_account: Account,
) -> None:
    patch_target = framework_config["platform_id_filter"]
    with patch(patch_target, MagicMock(return_value=True)):
        client = request.getfixturevalue(client_fixture)
        resp = client.get(f"/accounts/{other_account.id}")
        assert resp.status_code == 404


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_retrieve_resource_user_id_filter_required(
    client_fixture: str,
    framework_config: dict,
    request: pytest.FixtureRequest,
    other_account: Account,
) -> None:
    patch_target = framework_config["user_id_filter"]
    with patch(patch_target, MagicMock(return_value=True)):
        client = request.getfixturevalue(client_fixture)
        resp = client.get(f"/accounts/{other_account.id}")
        assert resp.status_code == 404


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_retrieve_resource_user_id_and_platform_id_filter_required(
    client_fixture: str,
    framework_config: dict,
    request: pytest.FixtureRequest,
    other_account: Account,
) -> None:
    platform_id_filter_target = framework_config["platform_id_filter"]
    user_id_filter_target = framework_config["user_id_filter"]

    with (
        patch(platform_id_filter_target, MagicMock(return_value=True)),
        patch(user_id_filter_target, MagicMock(return_value=True)),
    ):

        client = request.getfixturevalue(client_fixture)
        resp = client.get(f"/accounts/{other_account.id}")
        assert resp.status_code == 404


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_retrieve_resource_not_found(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.get('/accounts/unknown_id')
    assert resp.status_code == 404


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_update_resource_with_invalid_params(
    client_fixture: str, framework_config: dict, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    wrong_params = dict(wrong_param='wrong_value')
    response = client.patch(
        '/accounts/NOT_EXISTS',
        json=wrong_params,
    )
    assert response.status_code == framework_config['validation_error_code']


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_retrieve_custom_method(
    client_fixture: str, request: pytest.FixtureRequest, card: Card
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.get(f'/cards/{card.id}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert json_body['number'] == '*' * 16


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_update_resource_that_doesnt_exist(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.patch(
        '/accounts/5f9b4d0ff8d7255e3cc3c128',
        json=dict(name='Frida'),
    )
    assert resp.status_code == 404


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_update_resource(
    client_fixture: str, request: pytest.FixtureRequest, account: Account
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.patch(
        f'/accounts/{account.id}',
        json=dict(name='Maria Felix'),
    )
    json_body = resp.json()
    status_code = resp.status_code
    account.reload()
    assert json_body['name'] == 'Maria Felix'
    assert account.name == 'Maria Felix'
    assert status_code == 200


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_delete_resource(
    client_fixture: str, request: pytest.FixtureRequest, account: Account
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.delete(f'/accounts/{account.id}')
    json_body = resp.json()
    status_code = resp.status_code
    account.reload()
    assert status_code == 200
    assert json_body['deactivated_at'] is not None
    assert account.deactivated_at is not None


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_delete_resource_not_exists(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.delete('/accounts/1234')
    assert resp.status_code == 404


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
@pytest.mark.usefixtures('accounts')
def test_query_count_resource(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    query_params = dict(count=1, name='Frida Kahlo')
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert json_body['count'] == 1


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
@pytest.mark.usefixtures('accounts')
def test_query_all_with_limit(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    limit = 2
    query_params = dict(limit=limit)
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert len(json_body['items']) == limit
    assert json_body['next_page_uri'] is None


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
@pytest.mark.usefixtures('accounts')
def test_query_all_resource(
    client_fixture: str,
    request: pytest.FixtureRequest,
    accounts: list[Account],
) -> None:
    client = request.getfixturevalue(client_fixture)
    accounts = list(reversed(accounts))

    items = []
    page_uri = f'/accounts?{urlencode(dict(page_size=2))}'

    while page_uri:
        resp = client.get(page_uri)
        json_body = resp.json()
        status_code = resp.status_code
        assert status_code == 200
        items.extend(json_body['items'])
        page_uri = json_body['next_page_uri']

    assert len(items) == len(accounts)
    assert all(a.to_dict() == b for a, b in zip(accounts, items))


@pytest.mark.parametrize("client_fixture", ["chalice_client"])
def test_query_all_filter_active(
    client_fixture: str,
    request: pytest.FixtureRequest,
    account: Account,
    accounts: list[Account],
) -> None:
    client = request.getfixturevalue(client_fixture)
    query_params = dict(active=True)
    # Query active items
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    items = json_body['items']
    assert len(items) == len(accounts)
    assert all(item['deactivated_at'] is None for item in items)

    # Deactivate Item
    account.deactivated_at = dt.datetime.utcnow()
    account.save()
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    items = json_body['items']
    assert len(items) == len(accounts) - 1

    # Query deactivated items
    query_params = dict(active=False)
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    items = json_body['items']
    assert len(items) == 1
    assert items[0]['deactivated_at'] is not None


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_query_all_created_after(
    client_fixture: str,
    request: pytest.FixtureRequest,
    accounts: list[Account],
) -> None:
    client = request.getfixturevalue(client_fixture)
    created_at = dt.datetime(2020, 2, 1)
    expected_length = len([a for a in accounts if a.created_at > created_at])

    query_params = dict(created_after=created_at.isoformat())
    resp = client.get(f'/accounts?{urlencode(query_params)}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert len(json_body['items']) == expected_length


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_query_platform_id_filter_required(
    client_fixture: str,
    framework_config: dict,
    request: pytest.FixtureRequest,
    accounts: list[Account],
) -> None:
    client = request.getfixturevalue(client_fixture)
    patch_target = framework_config["platform_id_filter"]
    with patch(patch_target, MagicMock(return_value=True)):
        accounts = list(
            reversed(
                [
                    a
                    for a in accounts
                    if a.platform_id == TEST_DEFAULT_PLATFORM_ID
                ]
            )
        )

        items = []
        page_uri = f'/accounts?{urlencode(dict(page_size=2))}'

        while page_uri:
            resp = client.get(page_uri)
            json_body = resp.json()
            status_code = resp.status_code
            assert status_code == 200
            items.extend(json_body['items'])
            page_uri = json_body['next_page_uri']

        assert len(items) == len(accounts)
        assert all(a.to_dict() == b for a, b in zip(accounts, items))


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_query_user_id_filter_required(
    client_fixture: str,
    framework_config: dict,
    request: pytest.FixtureRequest,
    accounts: list[Account],
) -> None:
    client = request.getfixturevalue(client_fixture)
    patch_target = framework_config["user_id_filter"]
    with patch(patch_target, MagicMock(return_value=True)):
        accounts = list(
            reversed(
                [a for a in accounts if a.user_id == TEST_DEFAULT_USER_ID]
            )
        )
        items = []
        page_uri = f'/accounts?{urlencode(dict(page_size=2))}'

        while page_uri:
            resp = client.get(page_uri)
            json_body = resp.json()
            status_code = resp.status_code
            assert status_code == 200
            items.extend(json_body['items'])
            page_uri = json_body['next_page_uri']

        assert len(items) == len(accounts)
        assert all(a.to_dict() == b for a, b in zip(accounts, items))


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_query_resource_with_invalid_params(
    client_fixture: str, request: pytest.FixtureRequest, framework_config: dict
) -> None:
    client = request.getfixturevalue(client_fixture)
    wrong_params = dict(wrong_param='wrong_value')
    resp = client.get(f'/accounts?{urlencode(wrong_params)}')
    assert resp.status_code == framework_config['validation_error_code']


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
@pytest.mark.usefixtures('cards')
def test_query_custom_method(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    query_params = dict(page_size=2)
    resp = client.get(f'/cards?{urlencode(query_params)}')
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert len(json_body['items']) == 2
    assert all(card['number'] == '*' * 16 for card in json_body['items'])

    resp = client.get(json_body['next_page_uri'])
    json_body = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert len(json_body['items']) == 2
    assert all(card['number'] == '*' * 16 for card in json_body['items'])


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_cannot_create_resource(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.post('/billers', json=dict())
    assert resp.status_code == 405


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_cannot_query_resource(
    client_fixture: str, request: pytest.FixtureRequest, framework_config: dict
) -> None:
    client = request.getfixturevalue(client_fixture)
    query_params = dict(count=1, name='Frida Kahlo')
    resp = client.get(f'/transactions?{urlencode(query_params)}')
    assert resp.status_code == framework_config['method_not_allowed_code']


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_cannot_update_resource(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.patch('/transactions/123', json=dict())
    assert resp.status_code == 405


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_cannot_delete_resource(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.delete('/transactions/TR1234')
    assert resp.status_code == 405


@pytest.mark.parametrize(
    "client_fixture, framework_config",
    [
        ("fastapi_client", FRAMEWORK_CONFIGS["fastapi"]),
        ("chalice_client", FRAMEWORK_CONFIGS["chalice"]),
    ],
)
def test_not_found(
    client_fixture: str, request: pytest.FixtureRequest, framework_config: dict
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.get('/non-registered-endpoint')
    assert resp.status_code == framework_config['not_found_code']


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
def test_download_resource(
    client_fixture: str, request: pytest.FixtureRequest, file: File
) -> None:
    client = request.getfixturevalue(client_fixture)
    mimetype = 'application/pdf'
    resp = client.get(f'/files/{file.id}', headers={'Accept': mimetype})
    assert resp.status_code == 200
    assert resp.headers.get('Content-Type') == mimetype


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
@pytest.mark.usefixtures('users')
def test_filter_no_user_id_query(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.get(f'/users?platform_id={TEST_DEFAULT_PLATFORM_ID}')
    resp_json = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert len(resp_json['items']) == 1
    user1 = resp_json['items'][0]

    resp = client.get(f'/users?platform_id={TEST_SECOND_PLATFORM_ID}')
    resp_json = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert len(resp_json['items']) == 1
    user2 = resp_json['items'][0]
    assert user1['id'] != user2['id']


def test_update_user_with_ip(fastapi_client: TestClient, user: User) -> None:
    resp = fastapi_client.patch(
        f'/users/{user.id}', json={'name': 'Pedrito Sola'}
    )
    resp_json = resp.json()
    assert resp.status_code == 200
    assert resp_json['ip'] == 'testclient'
    assert resp_json['name'] == 'Pedrito Sola'


@pytest.mark.parametrize(
    "client_fixture", ["fastapi_client", "chalice_client"]
)
@pytest.mark.usefixtures('billers')
def test_filter_no_user_id_and_no_platform_id_query(
    client_fixture: str, request: pytest.FixtureRequest
) -> None:
    client = request.getfixturevalue(client_fixture)
    resp = client.get('/billers?name=ATT')
    resp_json = resp.json()
    status_code = resp.status_code
    assert status_code == 200
    assert len(resp_json['items']) == 1


def test_upload_resource(fastapi_client: TestClient) -> None:
    with TemporaryFile(mode='rb') as f:
        file_body = f.read()
    resp = fastapi_client.post(
        '/files',
        files=dict(file=(None, file_body), file_name=(None, 'test_file.txt')),
    )
    assert resp.status_code == 201
    json = resp.json()
    assert json['name'] == 'test_file.txt'


def test_upload_resource_with_invalid_form(fastapi_client: TestClient) -> None:
    wrong_form = dict(another_file=b'Whasaaaaap')
    resp = fastapi_client.post('/files', files=wrong_form)
    assert resp.status_code == 400
